#!/usr/bin/perl
# 
# Plug this into openvpn.conf as the client-connect script
# and dynamically add/remove iptables rules
#
# @author Joel Heenan 29/04/2010

use strict;
use warnings;
use Log::Log4perl qw(get_logger :levels);
use vars qw($logger);
use Readonly;
use IPTables::IPv4;
use Net::LDAP;

# *** evil constants ***

#logging
Readonly::Scalar our $script_log_file => '/var/log/openvpn_iptables.log';
Readonly::Scalar our $DEBUG           => 0;

#ldap
Readonly::Scalar our $LDAP_HOST       => '127.0.0.1';
Readonly::Scalar our $LDAP_BASE       => 'ou=RoutingGroups,ou=Groups,dc=EXAMPLE';
Readonly::Scalar our $LDAP_USERDN     => 'uid=%s,ou=People,dc=EXAMPLE';

#rules
Readonly::Scalar our $RULE_OUT_INTERFACE   => "eth0";
Readonly::Scalar our $RULE_JUMP            => "MASQUERADE";

Readonly::Array  our @DNS_SERVERS => qw( 127.0.0.1 );

sub initialize_logger($)
{
  my $debug = shift;

  my $level = 'INFO';

  if ($debug)
  {
    $level = 'DEBUG';
  }

  my $conf = qq(
    log4perl.rootLogger                = $level, Logfile, Screen

    log4perl.appender.Logfile          = Log::Log4perl::Appender::File
    log4perl.appender.Logfile.filename = $script_log_file
    log4perl.appender.Logfile.layout   = Log::Log4perl::Layout::PatternLayout
    log4perl.appender.Logfile.layout.ConversionPattern = %d %p> %F{1}:%L %M - %m%n
    log4perl.appender.Logfile.level    = DEBUG
    log4perl.appender.Logfile.mode     = append

    log4perl.appender.Screen         = Log::Log4perl::Appender::Screen
    log4perl.appender.Screen.layout  = Log::Log4perl::Layout::SimpleLayout
    log4perl.appender.Screen.level   = $level
  );


  Log::Log4perl::init( \$conf );

  $logger = get_logger();

  return;
}

sub connect_ldap
{
  my $ldap = Net::LDAP->new( $LDAP_HOST ) 
    or $logger->logdie("$@");
  $ldap->bind( protocol=> 3 );
  return $ldap;
}

# search through the routing groups, using the description as iptables
# rules (not including the source and the action, that is added later)
# return an array of all the descriptions
sub search_routing_groups($$)
{
  my ($ldap,$user) = @_;
  my @res = ();

  my $userdn = sprintf($LDAP_USERDN,$user);
  $logger->debug("Userdn: $userdn");

  my $ldapres = $ldap->search ( base   => "$LDAP_BASE",
                                scope  => "sub",
                                filter => "(&(objectclass=groupOfUniqueNames)(uniqueMember=$userdn))",
                                attrs  => [ "description" ],
                              );
  
  my $href = $ldapres->as_struct;

  foreach ( keys %$href )
  {
    $logger->debug( sprintf("Adding route %s:\n%s",$_,$$href{$_}{'description'}[0]) );
    push( @res, $$href{$_}{'description'}[0] );
  }

  return \@res;
}

sub add_iptables_rules($$)
{
  my ($rules,$remote_ip) = @_;

  my $natt = IPTables::IPv4::init('nat');

  $logger->debug("Cleaning up after ourselves");

  $natt->delete_entry("POSTROUTING", {
    "source"  => $remote_ip,
    "jump"    => $remote_ip,
  });

  if( $natt->is_chain($remote_ip) )
  {
    $natt->flush_entries($remote_ip);
    $natt->delete_chain($remote_ip)
      or $logger->logdie("Couldn't delete chain $remote_ip: $!");
  }

  $logger->debug("Adding new chain $remote_ip to the nat table");

  $natt->create_chain($remote_ip)
    or $logger->logdie("Couldn't create chain: $!");

  for my $rule ( @{ $rules } )
  {
    for my $line ( split("\n",$rule) )
    {
      my $destination = -1;
      my $dport = "";
      if( $line =~ /(-d|--destination) (\S+)/ )
      {
        $destination = $2;
      }
      else
      {
        $logger->logdie("Rule $rule must have a destination");
      }
      if( $line =~ /(--destination-port|--dport) (\S+)/ )
      {
        $dport = $2;
      }
      my %entryhash = (
        "source"        => $remote_ip,
        "destination"   => $destination,
        "protocol"      => "tcp",
        "out-interface" => $RULE_OUT_INTERFACE,
        "jump"          => $RULE_JUMP,
        );
      $entryhash{'destination-port'} = $dport if $dport;
      $logger->debug("Attempting to add new rule $destination $dport");
      $natt->append_entry($remote_ip, \%entryhash)
        or $logger->logdie("Could not append to chain $remote_ip, $!");
    }
  }

  for my $dns_server ( @DNS_SERVERS )
  {
    my %entryhash = (
        "source"           => $remote_ip,
        "destination"      => $dns_server,
        "destination-port" => "53",
        "protocol"         => "udp",
        "out-interface"    => $RULE_OUT_INTERFACE,
        "jump"             => $RULE_JUMP,
        );
    $natt->append_entry($remote_ip, \%entryhash)
      or $logger->logdie("Could not append to chain $remote_ip, $!");
  }

  $natt->append_entry("POSTROUTING", {
    "source"  => $remote_ip,
    "jump"    => $remote_ip,
  });

  $natt->commit()
    or $logger->logdie("Couldn't commit to table nat, $!");
}

sub main()
{
  my $user = $ARGV[0];
  my $remote_ip = $ARGV[1];
  initialize_logger($DEBUG);
  $logger->info("Connecting to LDAP $LDAP_HOST...");
  my $ldap = connect_ldap();
  $logger->info("Searching for routing group memberships for user $user");
  my $routing_groups = search_routing_groups($ldap,$user);
  $ldap->unbind;
  $logger->info("Adding iptables rules...");
  add_iptables_rules($routing_groups,$remote_ip);
}

main();

__END__
